'
' Interrupt-driven serial buffer for USART1 receive using a hardware high priority interrupt.
'
' This subroutine replaces the compiler's Hserin/Hrsin commands library subroutines
'
' For use with 18F devices only using the Proton+ Compiler version 3.3.0.0 Onwards
'
' Version 2 Written by Les Johnson 17th August 2007.
'
    #Disable HRSIn, HRSINTO, HRSIN_RCREG_READ           ' Disable the library subroutines for HRSIN with and without timeouts
'
' Create the SYSTEM variables and aliases for the Buffered commands
'
    Dim GEN     As Byte System                          ' \
    Dim GENH    As Byte System                          ' / Timeout value storage
    Dim PP0     As Byte System                          ' \
    Dim PP0H    As Byte System                          ' / Storage for FSR0L\H registers
    Dim PP1     As Byte System                          ' \
    Dim PP1H    As Byte System                          ' / Timeout inside loop counter
    Dim _USARTTimeoutInt As GEN
    Dim _USARTTimeoutIntH As GENH
    Dim _USARTTimeoutValue As _USARTTimeoutInt.Word     ' Alias the timeout value to GEN\H
    Dim _FSR0SaveInt As PP0
    Dim _FSR0SaveIntH As PP0H
    Dim _USART_FSR0_Save As _FSR0SaveInt.Word           ' Alias the FSR0L\H storage to PP0\H
    Dim InsideLoopInt As PP1
    Dim InsideLoopIntH As PP1H
    Dim _USART_InsideLoop As InsideLoopInt.Word         ' Alias the inside loop to PP1\H
    Dim _USART_OutsideLoop As PRODL.Word                ' Alias the outside loop to PRODL\H registers
    Dim _USART_FSR1_Save As Word System                 ' Storage for FSR1L\H registers
    Dim _USART_IndexIn As Byte System                   ' Pointer to the next empty location in the buffer
    Dim _USART_IndexOut As Byte System                  ' Pointer to the location of the oldest character in the buffer
'
' Create the actual serial buffer in high memory
'
Warnings = Off
Reminders = Off
    Reserve_RAM 100                                     ' Reserve some RAM at the top of memory as a serial buffer (MAX 255)
    Symbol _USART_BufferSize = __Reserve_Ram_Size      ' The amount of RAM reserved is the buffer size (MAX 255)
    Dim _USART_RingBuffer[_USART_BufferSize] As Byte At __Start_Of_Reserve_Ram    ' Array for holding received characters
Reminders = On
Warnings = On

    Dim _USART_FSR0 As FSR0L.Word
    Dim _USART_FSR1 As FSR1L.Word
'
' Point to the high priority hardware interrupt handler
'
    On_Hardware_Interrupt GoTo _Usart_High_Prior_Interrupt

    GoTo _Over_Usart_Handler                             ' Jump over the subroutines
'-------------------------------------------------------------------------------
' USART1 ring buffer interrupt handler
'
' Input     : Interrupt is triggered on reception of a byte from the USART
' Output    : Array _USART_RingBuffer holds the characters received
'           : Byte _USART_IndexIn points to the current location within the buffer
' Notes     : Does not indicate if an error occurs but does clear the error flags and carries on
'
_Usart_High_Prior_Interrupt:
    If PIR1bits_RCIF = 1 Then                            ' Was it a USART1 byte Receive that triggered the interrupt ?
        movlw 6                                          ' Yes. So Mask out unwanted bits
        andwf RCSTA,W                                    ' Check for errors
        bnz _Uart_Error                                  ' Was either error status bit set?
        _USART_FSR1_Save = _USART_FSR1                   ' Save FSR1L\H registers
        Inc _USART_IndexIn                               ' Move up the buffer
        If _USART_IndexIn >= _USART_BufferSize Then      ' End of buffer reached ?
           _USART_IndexIn = 0                            ' Yes. So clear _USART_IndexIn
        EndIf
        _USART_FSR1 = VarPtr _USART_RingBuffer           ' Point FSR1L\H to _USART_RingBuffer
        _USART_FSR1 = _USART_FSR1 + _USART_IndexIn       ' Add the buffer position to FSR1L\H
        INDF1 = RCREG                                    ' Place the received character into the buffer
        _USART_FSR1 = _USART_FSR1_Save                   ' Restore FSR1L\H registers
        retfie fast                                      ' Exit from the interrupt, restoring the WREG, STATUS, and BSR registers
_Uart_Error:
    Clear RCSTAbits_CREN                                 ' Clear receiver status
    Set RCSTAbits_CREN
    Else                                                 ' Otherwise...
' <<< Other Interrupt Conditions and Code here >>>
    EndIf
    retfie fast                                          ' Exit from the interrupt, restoring the WREG, STATUS, and BSR registers
'------------------------------------------------------------------------------
' Wait for a character from the interrupt driven circular buffer with timeout
'
' Input     : GEN\GENH hold the timeout value in approx ms (0 to 65535)
' Output    : PP0 and WREG hold the value received
'           : CARRY Flag (STATUS.0) Clear if timeout out
' Notes     : Uses PRODL\H as temporary variable storage
'           : Uses FSR0L\H as buffer the pointer
'
#ifdef Hrsinto#Req
HRSIn2:
@Read#Buffer_Req = 1                                 ; Enable the _USART_GET_CHAR subroutine
    _USART_OutsideLoop = _USARTTimeoutValue          ' Save the timeout value so it doesn't get overwritten
    Clear _USART_InsideLoop                          ' Reset the inside loop counter
_Usart_Getbuf_Outside:
    bra     $ + 2                                    ' Delay for 2 cycles within the outside loop
_Usart_Getbuf_Inside:
    #ifdef Watchdog_Req                              ' Delay for 1 cycle within the inside loop
        clrwdt
    #else
        nop
    #endif
    movf      _USART_IndexIn,W                       ' \
    subwf     _USART_IndexOut,W                      ' / Is there anything in the serial buffer ?
    bnz       _Usart_Get_Char                        ' Yes. So fetch it
    WREG = 255                                       ' \
    addwf     _USART_InsideLoop,F                    '  \
    addwfc    _USART_InsideLoopH,F                   '    Decrement the inside and outside loops
    addwfc    _USART_OutsideLoop,F                   '  /
    addwfc    _USART_OutsideLoopH,F                  ' /
    btfss     STATUS,C
    ret                                              ' Return with the CARRY flag clear to indicate timed out
    infsnz    _USART_InsideLoop,W
    incfsz    _USART_InsideLoopH,W
    bra       _Usart_Getbuf_Outside
    _USART_InsideLoop = ((59 * Xtal) / 4)            ' Set the inside loop counter based upon the XTAL frequency
    bra       _Usart_Getbuf_Inside
#endif
'------------------------------------------------------------------------------
' Wait for a character from the interrupt driven circular buffer without timeout
'
' Input     : None
' Output    : WREG holds the value received
' Notes     : Uses FSR0L\H as buffer pointers
'
#ifdef Hrsinx#Req
@HRSIN
@Read#Buffer_Req = 1                                  ; Enable the _USART_GET_CHAR subroutine
    While _USART_IndexIn = _USART_IndexOut            ' Wait for a character to apear in the serial buffer
    #ifdef Watchdog_Req
        clrwdt                                        ' Clear the watchdog timer within the loop if required
    #endif
    Wend
#endif
' Fall through to _USART_GET_CHAR
'------------------------------------------------------------------------------
' Read a character from the interrupt driven circular buffer
'
' Input     : None
' Output    : WREG holds the value received
' Notes     : Uses FSR0L\H as buffer pointers
'
#ifdef Read#Buffer_Req
_Usart_Get_Char:
    Inc _USART_IndexOut                               ' Increment _USART_IndexOut pointer (0 to 255)
    If _USART_IndexOut >= _USART_BufferSize Then      ' End of buffer reached ?
        _USART_IndexOut = 0                           ' Yes. So clear _USART_IndexOut.
    EndIf
    _USART_FSR0_Save = _USART_FSR0                    ' Save FSR0L\H registers
    _USART_FSR0 = VarPtr _USART_RingBuffer            ' Point FSR0L\H to _USART_RingBuffer
    _USART_FSR0 = _USART_FSR0 + _USART_IndexOut       ' Add the buffer position to FSR0L\H
    WREG = INDF0                                      ' Read buffer location (_USART_IndexOut) into WREG
    PP0 = WREG                                        ' Also place it into PP0
    _USART_FSR0 = _USART_FSR0_Save                    ' Restore FSR0\H registers
    bsf STATUS,C                                      ' Set the CARRY flag to indicate a byte received
    Return
#endif
'--------------------------------------------------------------------------------
' Initialise the USART1 interrupt
'
' Input     : None
' Output    : None
' Notes     : Enables interrupt on USART1 receive.
'           : Enables global and peripheral interrupts
'           : If Prioritised interrupts are used, the USART interrupt priority is made high
'
Init_Usart_Interrupt Macro
    GoSub _Init_Usart
    Endm
#ifdef Init_Usart_Interrupt#Req
_Init_Usart:
    _USART_IndexIn = 0                                ' Clear the buffer internal pointer
    _USART_IndexOut = 0                               ' Clear the buffer external pointer
    PIE1bits_RCIE = 1                                 ' Enable interrupt on USART1 receive
    #ifdef __Low_Interrupts_Enabled                   ' Are we using low priority interrupts as well ?
        IPR1bits_RCIP = 1                             ' Yes. So USART1 Receive Interrupt to High priority
    #endif
    INTCONbits.GIE = 1
    INTCONbits_PEIE = 1                               ' Enable global and peripheral interrupts
    Return
#endif
'--------------------------------------------------------------------------------
' Clear the Serial Buffer
'
' Input     : None
' Output    : None
' Notes     : Also resets the index pointers to the serial buffer
'
Clear_Serial_Buffer Macro
    GoSub _Clear_Usart_Buffer
    Endm
#ifdef Clear_Serial_Buffer#Req
_Clear_Usart_Buffer:
    PIE1bits_RCIE = 0                                 ' Disable interrupt on USART1 receive
    Clear _USART_RingBuffer                           ' Clear the serial buffer
    _USART_IndexIn = 0                                ' Clear the buffer internal pointer
    _USART_IndexOut = 0                               ' Clear the buffer external pointer
    PIE1bits_RCIE = 1                                 ' Re-Enable interrupt on USART1 receive
    Return
#endif
'--------------------------------------------------------------------------------
_Over_Usart_Handler:


